"""A collection of helper functions for the map sub-package."""
import json
from pathlib import Path
import numpy as np


def binary_mtx_dimension(filepath):
    """Return binary Radiance matrix dimensions if exist.

    This function returns NROWS, NCOLS, NCOMP and number of header lines including the
    white line after last header line.

    Args:
        filepath: Full path to Radiance file.

    Returns:
        nrows, ncols, ncomp, line_count, fmt
    """
    try:
        inf = open(filepath, 'rb', encoding='utf-8')
    except Exception:
        inf = open(filepath, 'rb')
    try:
        first_line = next(inf).rstrip().decode('utf-8')
        if first_line[:10] != '#?RADIANCE':
            error_message = (
                f'File with Radiance header must start with #?RADIANCE not '
                f'{first_line}.'
            )
            raise ValueError(error_message)

        header_lines = [first_line]
        nrows = ncols = ncomp = None
        for line in inf:
            line = line.rstrip().decode('utf-8')
            header_lines.append(line)
            if line[:6] == 'NROWS=':
                nrows = int(line.split('=')[-1])
            if line[:6] == 'NCOLS=':
                ncols = int(line.split('=')[-1])
            if line[:6] == 'NCOMP=':
                ncomp = int(line.split('=')[-1])
            if line[:7] == 'FORMAT=':
                fmt = line.split('=')[-1]
                break

        if not nrows or not ncols:
            error_message = (
                f'NROWS or NCOLS was not found in the Radiance header. NROWS '
                f'is {nrows} and NCOLS is {ncols}. The header must have both '
                f'elements.'
            )
            raise ValueError(error_message)
        return nrows, ncols, ncomp, len(header_lines) + 1, fmt
    finally:
        inf.close()


def binary_to_array(
        binary_file, nrows=None, ncols=None, ncomp=None, fmt=None,
        line_count=0):
    """Read a Radiance binary file as a NumPy array.

    Args:
        binary_file: Path to binary Radiance file.
        nrows: Number of rows in the Radiance file.
        ncols: Number of columns in the Radiance file.
        ncomp: Number of components of each element in the Radiance file.
        fmt: Format of the Radiance file. Can be either "ascii", "float", or "double.
        line_count: Number of lines to skip in the input file. Usually used to
            skip the header.

    Returns:
        A NumPy array.
    """
    if (nrows or ncols or ncomp or fmt) is None:
        # get nrows, ncols and header line count
        nrows, ncols, ncomp, line_count, fmt = binary_mtx_dimension(binary_file)
    with open(binary_file, 'rb') as reader:
        # skip first n lines from reader
        for i in range(line_count):
            reader.readline()

        if fmt == 'ascii':
            array = np.loadtxt(reader, dtype=np.float32)
        elif fmt == 'float':
            array = np.fromfile(reader, dtype=np.float32)
        elif fmt == 'double':
            array = np.fromfile(reader, dtype=np.float64)

        if ncomp != 1:
            array = array.reshape(nrows, ncols, ncomp)
        else:
            array = array.reshape(nrows, ncols)

    return array


def load_matrix(matrix_file, delimiter=','):
    with open(matrix_file, 'rb') as inf:
        first_char = inf.read(1)
        second_char = inf.read(1)
    is_text = True if first_char.isdigit() or second_char.isdigit() else False
    if is_text:
        array = np.genfromtxt(
            matrix_file, delimiter=delimiter, encoding='utf-8',
            filling_values=np.nan)
        if array.ndim == 1:
            array = array.reshape(-1, 1)
        if np.isnan(array[:, -1]).all():
            # remove last column if all in column is NaN
            # this may happen if the CSV has trailing commas
            array = array[:, :-1]
    else:
        array = np.load(matrix_file)

    return array


def restore_original_distribution(
        input_folder, output_folder, extension='npy', dist_info=None,
        output_extension='ill', as_text=False, fmt='%.2f', input_delimiter=',',
        delimiter='tab'):
    """Restructure files to the original distribution based on the distribution info.
    
    It will assume that the files in the input folder are NumPy files. However,
    if it fails to load the files as arrays it will try to load from binary
    Radiance files to array.

    Args:
        input_folder: Path to input folder.
        output_folder: Path to the new restructured folder
        extension: Extension of the files to collect data from. Default is ``npy`` for
            NumPy files. Another common extension is ``ill`` for the results of daylight
            studies.
        dist_info: Path to dist_info.json file. If None, the function will try to load
            ``_redist_info.json`` file from inside the input_folder. (Default: None).
        output_extension: Output file extension. This is only used if as_text
            is set to True. Otherwise the output extension will be ```npy``.
        as_text: Set to True if the output files should be saved as text instead
            of NumPy files.
        fmt: Format for the output files when saved as text.
        input_delimiter: Delimiter for the input files. This is used only if the
            input files are text files.
        delimiter: Delimiter for the output files when saved as text.
    """
    if not dist_info:
        _redist_info_file = Path(input_folder, '_redist_info.json')
    else:
        _redist_info_file = Path(dist_info)

    assert _redist_info_file.is_file(), 'Failed to find %s' % _redist_info_file

    with open(_redist_info_file) as inf:
        data = json.load(inf)

    # create output folder
    output_folder = Path(output_folder)
    if not output_folder.is_dir():
        output_folder.mkdir(parents=True, exist_ok=True)

    src_file = Path()
    for f in data:
        output_file = Path(output_folder, f['identifier'])
        # ensure the new folder is created. in case the identifier has a subfolder
        parent_folder = output_file.parent
        if not parent_folder.is_dir():
            parent_folder.mkdir()

        out_arrays = []
        for src_info in f['dist_info']:
            st = src_info['st_ln']
            end = src_info['end_ln']
            new_file = Path(input_folder, '%s.%s' % (src_info['identifier'], extension))
            if not new_file.samefile(src_file):
                src_file = new_file
                try:
                    array = np.load(src_file)
                except:
                    try:
                        array = binary_to_array(src_file)
                    except:
                        try:
                            array = np.loadtxt(
                                src_file, delimiter=input_delimiter)
                        except Exception:
                            raise RuntimeError(
                                f'Failed to load input file "{src_file}"')
            slice_array = array[st:end+1,:]

            out_arrays.append(slice_array)

        out_array = np.concatenate(out_arrays)
        # save numpy array, .npy extension is added automatically
        if not as_text:
            np.save(output_file, out_array)
        else:
            if output_extension.startswith('.'):
                output_extension = output_extension[1:]
            if delimiter == 'tab':
                delimiter = '\t'
            elif delimiter == 'space':
                delimiter = ' '
            elif delimiter == 'comma':
                delimiter = ','
            np.savetxt(output_file.with_suffix(f'.{output_extension}'),
                       out_array, fmt=fmt, delimiter=delimiter)



# coding=utf-8
"""Parameters for specifying body characteristics for the PET model."""
from __future__ import division
import re

from ._base import ComfortParameter


class PETParameter(ComfortParameter):
    """Parameters specifying body characteristics for the PET model.

    Args:
        age: The age of the human subject in years. (Default: 36 years for middle age
            of the average worldwide life expectancy).
        sex: A value between 0 and 1 to indicate the sex of the human subject,
            which influences the computation of basal metabolism. 0 indicates male.
            1 indicates female and any number in between denotes a weighted average
            between the two. (Default: 0.5).
        height: The height of the human subject in meters. Average male height
            is around 1.75m while average female height is 1.55m. (Default: 1.65m
            for a worldwide average between male and female height).
        body_mass: The body mass of the human subject in kilograms. (Default: 62 kg
            for the worldwide average adult human body mass).
        posture: A text string indicating the posture of the body. Letters must
            be lowercase. Default is "standing". Choose from the following:

            * standing
            * seated
            * crouching

        humid_acclimated: A boolean to note whether the human subject is acclimated
            to a humid/tropical climate (True) or is acclimated to a temperate
            climate (False). When True, the categories developed by Lin and
            Matzarakis (2008) will be used to assess comfort instead of the original
            categories developed by Matzarakis and Mayer (1996).

    Properties:
        * age
        * sex
        * height
        * body_mass
        * posture
        * humid_acclimated
    """
    _model = 'Physiological Equivalent Temperature'
    POSTURES = ('standing', 'seated', 'crouching')
    __slots__ = ('_age', '_sex', '_height', '_body_mass', '_posture',
                 '_humid_acclimated')

    def __init__(self, age=None, sex=None, height=None, body_mass=None,
                 posture=None, humid_acclimated=False):
        """Initialize PET Body Parameters."""
        if age is not None:
            assert 0 <= age <= 120, 'Age must be between 0 and 120. Got {}'.format(age)
            self._age = age
        else:
            self._age = 36

        if sex is not None:
            assert 0 <= sex <= 1, 'Sex must be between 0 and 1. Got {}'.format(sex)
            self._sex = sex
        else:
            self._sex = 0.5

        if height is not None:
            assert 0.6 <= height <= 3, \
                'Height must be between 0.6 and 3. Got {}'.format(height)
            self._height = height
        else:
            self._height = 1.65

        if body_mass is not None:
            assert 0 < body_mass <= 300, 'Body mass must be between'\
                ' 0 and 300. Got {}'.format(body_mass)
            self._body_mass = body_mass
        else:
            self._body_mass = 62

        if posture is not None:
            assert isinstance(posture, str), 'posture must be a string.'\
                ' Got {}'.format(type(posture))
            assert posture.lower() in self.POSTURES, 'posture {} is not '\
                'acceptable. Choose from {}'.format(posture, self.POSTURES)
            self._posture = posture
        else:
            self._posture = 'standing'

        self._humid_acclimated = bool(humid_acclimated)

    @classmethod
    def from_dict(cls, data):
        """Create a PETParameter object from a dictionary.

        Args:
            data: A PETParameter dictionary in following the format below.

        .. code-block:: python

            {
            'type': 'PETParameter',
            'age': 36,
            'sex': 1,
            'height': 1.55,
            'body_mass': 45,
            'posture': 'standing',
            'humid_acclimated': True
            }
        """
        assert data['type'] == 'PETParameter', \
            'Expected PETParameter dictionary. Got {}.'.format(data['type'])
        age = data['age'] if 'age' in data else None
        sex = data['sex'] if 'sex' in data else None
        height = data['height'] if 'height' in data else None
        body_mass = data['body_mass'] if 'body_mass' in data else None
        posture = data['posture'] if 'posture' in data else None
        h_acc = data['humid_acclimated'] if 'humid_acclimated' in data else False
        return cls(age, sex, height, body_mass, posture, h_acc)

    @classmethod
    def from_string(cls, pet_parameter_string):
        """Create an PETParameter object from an PETParameter string."""
        str_pattern = re.compile(r"\-\-(\S*\s\S*)")
        matches = str_pattern.findall(pet_parameter_string)
        par_dict = {item.split(' ')[0]: item.split(' ')[1] for item in matches}

        age = float(par_dict['age']) if 'age' in par_dict else None
        sex = float(par_dict['sex']) if 'sex' in par_dict else None
        height = float(par_dict['height']) if 'height' in par_dict else None
        body_mass = float(par_dict['body-mass']) if 'body-mass' in par_dict else None
        posture = par_dict['posture'] if 'posture' in par_dict else None
        h_acc = False if 'acclimated' not in par_dict \
            or par_dict['acclimated'].lower() == 'temperate' else True
        return cls(age, sex, height, body_mass, posture, h_acc)

    @property
    def age(self):
        """The age of the human subject in years."""
        return self._age

    @property
    def sex(self):
        """A value between 0 and 1 to indicate the sex of the human subject.

        This influences the computation of basal metabolism.
        0 indicates male.
        1 indicates female.
        Any number in between denotes a weighted average between the two.
        """
        return self._sex

    @property
    def height(self):
        """The height of the human subject in meters.

        Average male height is around 1.75m while average female height is 1.55m.
        """
        return self._height

    @property
    def body_mass(self):
        """The body mass of the human subject in kilograms."""
        return self._body_mass

    @property
    def posture(self):
        """A text string indicating the posture of the body."""
        return self._posture

    @property
    def humid_acclimated(self):
        """A boolean for whether subject is acclimated to a humid or temperate climate.

        When True, the categories developed by Lin and Matzarakis (2008) for a
        humid/subtropical climate will be used to assess comfort instead of the
        original categories developed by Matzarakis and Mayer (1996).
        """
        return self._humid_acclimated

    def to_dict(self):
        """PETParameter dictionary representation."""
        return {
            'type': 'PETParameter',
            'age': self.age,
            'sex': self.sex,
            'height': self.height,
            'body_mass': self.body_mass,
            'posture': self.posture,
            'humid_acclimated': self.humid_acclimated
        }

    def __copy__(self):
        return PETParameter(
            self.age, self.sex, self.height, self.body_mass, self.posture,
            self.humid_acclimated)

    def __repr__(self):
        """PET body parameters representation."""
        base = '--age {} --sex {} --height {} --body-mass {} --posture {}'.format(
            self.age, self.sex, self.height, self.body_mass, self.posture)
        if self.humid_acclimated:
            base = '{} --acclimated humid'.format(base)
        return base



# coding=utf-8
"""Object for plotting an UTCI comfort polygon on a Psychrometric Chart."""
from __future__ import division

from ladybug_geometry.geometry2d.pointvector import Point2D, Vector2D
from ladybug_geometry.geometry2d.ray import Ray2D
from ladybug_geometry.geometry2d.line import LineSegment2D
from ladybug_geometry.geometry2d.polyline import Polyline2D

from ladybug.psychchart import PsychrometricChart
from ladybug.psychrometrics import humid_ratio_from_db_rh
from ladybug._datacollectionbase import BaseCollection
from ladybug.datacollection import HourlyContinuousCollection, \
    HourlyDiscontinuousCollection
from ladybug.datatype.temperature import Temperature
from ladybug.datatype.temperaturedelta import TemperatureDelta
from ladybug.datatype.thermalcondition import ThermalComfort
from ladybug.datatype.generic import GenericType

from ..utci import calc_missing_utci_input
from ..parameter.utci import UTCIParameter


class PolygonUTCI(object):
    """Object to plot an UTCI comfort polygon on a Psychrometric Chart.

    Args:
        psychrometric_chart: A ladybug-core PsychrometricChart object on which the
            UTCI comfort polygon will be plot.
        rad_temperature: A list of numbers for the mean radiant temperature in Celsius.
            If None, a polygon for operative temperature will be plot, assuming that
            radiant temperature and air temperature are the same. (Default: None).
        wind_speed: A list of numbers for the meteorological wind speed values in m/s
            (measured 10 m above the ground). If None, this will default to a low
            wind speed of 0.5 m/s, which is the lowest input speed that is
            recommended for the UTCI model.
        comfort_parameter: Optional UTCIParameter object to specify parameters under
            which conditions are considered acceptable. If None, default will
            assume comfort thresholds consistent with those used by meteorologists
            to categorize outdoor conditions.

    Properties:
        * psychrometric_chart
        * rad_temperature
        * wind_speed
        * comfort_parameter
        * polygon_count
        * left_comfort_lines
        * right_comfort_lines
        * left_comfort_line
        * right_comfort_line
        * comfort_polygons
        * merged_comfort_polygon
        * comfort_values
        * comfort_data
        * merged_comfort_values
        * merged_comfort_data
        * is_comfort_too_hot
        * is_comfort_too_cold
        * very_strong_cold_polygon
        * strong_cold_polygon
        * moderate_cold_polygon
        * moderate_heat_polygon
        * strong_heat_polygon
        * very_strong_heat_polygon
    """
    TEMP_TYPE = Temperature()
    DELTA_TEMP_TYPE = TemperatureDelta()
    POLYGON_INCLUSION_TYPE = GenericType(
        'Polygon Inclusion', 'status', 0, 1, unit_descr={0: 'Outside', 1: 'Inside'})

    def __init__(self, psychrometric_chart, rad_temperature=None, wind_speed=None,
                 comfort_parameter=None):
        """Initialize a UTCI comfort polygon."""
        # check the psychrometric_chart input
        assert isinstance(psychrometric_chart, PsychrometricChart), 'PolygonUTCI ' \
            'psychrometric_chart must be a ladybug PsychrometricChart. ' \
            'Got {}.'.format(type(psychrometric_chart))
        self._psychrometric_chart = psychrometric_chart

        # determine the number of comfort polygons to be drawn
        all_data = (rad_temperature, wind_speed)
        param_lens = [len(arr) for arr in all_data if arr is not None]
        self._polygon_count = max(param_lens) if len(param_lens) != 0 else 0
        self._polygon_count = 1 if self._polygon_count == 0 else self._polygon_count

        # check parameters with defaults
        self._rad_temperature = self._check_input(
            rad_temperature, 'rad_temperature', None)
        self._wind_speed = self._check_input(wind_speed, 'wind_speed', 0.5, True)

        # check comfort parameters
        if comfort_parameter is None:
            self._comfort_par = UTCIParameter()
        else:
            assert isinstance(comfort_parameter, UTCIParameter), 'comfort_parameter '\
                'must be a UTCIParameter object. Got {}'.format(type(comfort_parameter))
            self._comfort_par = comfort_parameter

        # create the left and right polylines
        _left, _right = [], []
        for p in range(self._polygon_count):
            min_poly, max_poly = self.comfort_polylines(p)
            _left.append(min_poly)
            _right.append(max_poly)
        self._left_comfort_lines, self._right_comfort_lines = tuple(_left), tuple(_right)

        # set parameters to None, which will be computed on demand
        self._left_comfort_line = None
        self._right_comfort_line = None
        self._comfort_polygons = None
        self._merged_comfort_polygons = None
        self._comfort_values = None
        self._comfort_data = None
        self._merged_comfort_values = None
        self._merged_comfort_data = None

    @property
    def psychrometric_chart(self):
        """The ladybug PsychrometricChart object on which the polygons are plot."""
        return self._psychrometric_chart

    @property
    def rad_temperature(self):
        """Tuple of mean radiant temperature (MRT) values in degrees C.

        None indicates that the radiant temperature is the same as the air temperature.
        """
        return self._rad_temperature

    @property
    def wind_speed(self):
        """Tuple of meteorological wind speed values in m/s."""
        return self._wind_speed

    @property
    def comfort_parameter(self):
        """UTCI comfort parameters that are assigned to this object."""
        return self._comfort_par

    @property
    def polygon_count(self):
        """Integer for the number of comfort polygons contained on the object."""
        return self._polygon_count

    @property
    def left_comfort_lines(self):
        """Tuple of Polyline2D for the left of the comfort polygons."""
        return self._left_comfort_lines

    @property
    def right_comfort_lines(self):
        """Tuple of Polyline2D for the right of the comfort polygons."""
        return self._right_comfort_lines

    @property
    def left_comfort_line(self):
        """A single Polyline2D for the left of the merged comfort polygons."""
        if self._left_comfort_line is None:
            li = self._left_comfort_lines
            self._left_comfort_line = li[0] if len(li) == 1 else self._min_polylines(li)
        return self._left_comfort_line

    @property
    def right_comfort_line(self):
        """A single Polyline2D for the right of the merged comfort polygons."""
        if self._right_comfort_line is None:
            li = self._right_comfort_lines
            self._right_comfort_line = li[0] if len(li) == 1 else self._max_polylines(li)
        return self._right_comfort_line

    @property
    def comfort_polygons(self):
        """A tuple of tuples where each sub-tuple defines one comfort polygon.

        Sub-tuple comfort polygons consist of four or five Polyline2D or LineSegment2D
        that are ordered as follows (left, bottom, right, top).
        """
        if self._comfort_polygons is None:
            self._comfort_polygons = []
            ll, rl = self.left_comfort_lines, self.right_comfort_lines
            for lt, rt in zip(ll, rl):
                self._comfort_polygons.append(self._build_comfort_polygon(lt, rt))
        return tuple(self._comfort_polygons)

    @property
    def merged_comfort_polygon(self):
        """A tuple of Polyline2D or LineSegment2D that define the merged comfort polygon.

        Comfort polygon consists of four or five Polyline2D or LineSegment2D that
        are ordered as follows (left, bottom, right, top).
        """
        if self._merged_comfort_polygons is None:
            lt, rt = self.left_comfort_line, self.right_comfort_line
            self._merged_comfort_polygons = self._build_comfort_polygon(lt, rt)
        return self._merged_comfort_polygons

    @property
    def comfort_values(self):
        """A tuple of tuples with each sub-tuple representing one of comfort polygons.

        Each sub-tuple contains 0/1 values for whether the point is inside the
        comfort polygon or not.
        """
        if self._comfort_values is None:
            self._comfort_values = []
            for poly in self.comfort_polygons:
                self._comfort_values.append(self._evaluate_comfort(poly[0], poly[2]))
        return tuple(self._comfort_values)

    @property
    def comfort_data(self):
        """A tuple of data collections or 0/1 values for each of the comfort polygons."""
        if self._comfort_data is None:
            self._comfort_data = []
            for i, dat in enumerate(self.comfort_values):
                if len(dat) == 1:
                    self._comfort_data.append(dat[0])
                else:
                    name = 'Comfort {}'.format(i + 1)
                    self._comfort_data.append(self.create_collection(dat, name))
        return tuple(self._comfort_data)

    @property
    def merged_comfort_values(self):
        """A tuple of 0/1 for whether each point is in the merged comfort polygon or not.
        """
        if self._merged_comfort_values is None:
            poly = self.merged_comfort_polygon
            self._merged_comfort_values = self._evaluate_comfort(poly[0], poly[2])
        return self._merged_comfort_values

    @property
    def merged_comfort_data(self):
        """A data collection or 0/1 for whether the data is in merged comfort polygon.
        """
        if self._merged_comfort_data is None:
            if len(self.merged_comfort_values) == 1:
                self._merged_comfort_data = self.merged_comfort_values[0]
            else:
                self._merged_comfort_data = \
                    self.create_collection(self.merged_comfort_values, 'Comfort')
        return self._merged_comfort_data

    @property
    def is_comfort_too_hot(self):
        """Boolean to note whether comfort polygons are off the chart on the hot side."""
        psy = self.psychrometric_chart
        return self.merged_comfort_polygon[2][0].x >= psy.base_point.x + \
            (psy._max_temperature - psy._min_temperature) * psy._x_dim

    @property
    def is_comfort_too_cold(self):
        """Boolean to note whether comfort polygons are off the chart on the cold side.
        """
        psy = self.psychrometric_chart
        return self.merged_comfort_polygon[0][0].x <= psy.base_point.x

    @property
    def very_strong_cold_polygon(self):
        """A tuple of Polyline2D or LineSegment2D for the very strong cold polygon."""
        left_line = self.stress_polyline(self._comfort_par.very_strong_cold_thresh)
        right_line = self.stress_polyline(self._comfort_par.strong_cold_thresh)
        return self._build_comfort_polygon(left_line, right_line)
    
    @property
    def strong_cold_polygon(self):
        """A tuple of Polyline2D or LineSegment2D for the strong cold polygon."""
        left_line = self.stress_polyline(self._comfort_par.strong_cold_thresh)
        right_line = self.stress_polyline(self._comfort_par.moderate_cold_thresh)
        return self._build_comfort_polygon(left_line, right_line)

    @property
    def moderate_cold_polygon(self):
        """A tuple of Polyline2D or LineSegment2D for the moderate cold polygon."""
        left_line = self.stress_polyline(self._comfort_par.moderate_cold_thresh)
        right_line = self.left_comfort_line
        return self._build_comfort_polygon(left_line, right_line)
    
    @property
    def moderate_heat_polygon(self):
        """A tuple of Polyline2D or LineSegment2D for the moderate heat polygon."""
        left_line = self.right_comfort_line
        right_line = self.stress_polyline(self._comfort_par.moderate_heat_thresh)
        return self._build_comfort_polygon(left_line, right_line)

    @property
    def strong_heat_polygon(self):
        """A tuple of Polyline2D or LineSegment2D for the strong heat polygon."""
        left_line = self.stress_polyline(self._comfort_par.moderate_heat_thresh)
        right_line = self.stress_polyline(self._comfort_par.strong_heat_thresh)
        return self._build_comfort_polygon(left_line, right_line)
    
    @property
    def very_strong_heat_polygon(self):
        """A tuple of Polyline2D or LineSegment2D for the very strong heat polygon."""
        left_line = self.stress_polyline(self._comfort_par.strong_heat_thresh)
        right_line = self.stress_polyline(self._comfort_par.very_strong_heat_thresh)
        return self._build_comfort_polygon(left_line, right_line)

    def comfort_polylines(self, polygon_index):
        """Get the left and right Polyline2D that define a UTCI polygon comfort range.

        Args:
            polygon_index: Integer for the comfort polygon for which min and max
                temperature will be computed.

        Returns:
            The left and right Polyline2D that define the comfort range.
        """
        # get the air temperature and humidity rations
        rel_humids = (0, 20, 40, 60, 80, 100)
        pres = self.psychrometric_chart.average_pressure
        air_temps = self.max_min_air_temperatures(polygon_index, rel_humids)
        humid_ratios = []
        for i, temp in enumerate(air_temps):
            hr_min = humid_ratio_from_db_rh(temp[0], rel_humids[i], pres)
            hr_max = humid_ratio_from_db_rh(temp[1], rel_humids[i], pres)
            humid_ratios.append((hr_min, hr_max))

        # create the points from the temperature and humidity ratios
        psy, left_pts, right_pts = self.psychrometric_chart, [], []
        for hr, ta in zip(humid_ratios, air_temps):
            ta1, ta2 = ta if not psy.use_ip else self.TEMP_TYPE.to_unit(ta, 'F', 'C')
            left_pts.append(Point2D(psy.t_x_value(ta1), psy.hr_y_value(hr[0])))
            right_pts.append(Point2D(psy.t_x_value(ta2), psy.hr_y_value(hr[1])))
        return Polyline2D(left_pts, interpolated=True), \
            Polyline2D(right_pts, interpolated=True)

    def stress_polyline(self, stress_temperature):
        """Get a Polyline2D that defines a specific UTCI value,.
        
        Used to construct stress polygons.

        Args:
            stress_temperature: Number in degrees Celsius for the UTCI value of the
                given stress threshold line to be computed.

        Returns:
            A Polyline2D that define the comfort range.
        """
        # determine which comfort polygon conditions to use
        if len(self.left_comfort_lines) == 1:
            polygon_index = 0
        elif stress_temperature < self._comfort_par.cold_thresh:
            polygon_index = self._min_index(self.left_comfort_lines)
        elif stress_temperature > self._comfort_par.heat_thresh:
            polygon_index = self._max_index(self.right_comfort_lines)
        else:
            polygon_index = 0

        # get the air temperature and humidity rations
        rel_humids = (0, 20, 40, 60, 80, 100)
        pres = self.psychrometric_chart.average_pressure
        utci_dict = self._utci_dict(polygon_index)
        air_temps = []
        for rh in rel_humids:
            utci_dict['rh'] = rh
            t_dict = calc_missing_utci_input(stress_temperature, utci_dict)
            air_temps.append(t_dict['ta'])
        humid_ratios = []
        for i, temp in enumerate(air_temps):
            hr_val = humid_ratio_from_db_rh(temp, rel_humids[i], pres)
            humid_ratios.append(hr_val)

        # create the points from the temperature and humidity ratios
        psy, stress_pts = self.psychrometric_chart, []
        for hr, ta in zip(humid_ratios, air_temps):
            if psy.use_ip:
                ta = self.TEMP_TYPE.to_unit([ta], 'F', 'C')[0]
            stress_pts.append(Point2D(psy.t_x_value(ta), psy.hr_y_value(hr)))
        return Polyline2D(stress_pts, interpolated=True)

    def max_min_air_temperatures(self, polygon_index, rel_humid):
        """Get the max and min air temperature for a comfort polygon at a relative humid.

        Args:
            polygon_index: Integer for the comfort polygon for which min and max
                temperature will be computed.
            rel_humid: A list of relative humidity values for which air temperature
                will be computed.

        Returns:
            A list of tuples where each tuple contains two air temperature values in
            Celsius. The first air temperature is the minimum temperature that meets
            the PPD threshold. The second is the maximum that meets the PPD
            threshold
        """
        # get the UTCI thresholds and UTCI dict
        utci_min = self._comfort_par.cold_thresh
        utci_max = self._comfort_par.heat_thresh
        utci_dict = self._utci_dict(polygon_index)

        # compute the min and max air temperatures of relative humidity
        air_temperatures = []
        for rh in rel_humid:
            utci_dict['rh'] = rh
            min_dict = calc_missing_utci_input(utci_min, utci_dict)
            max_dict = calc_missing_utci_input(utci_max, utci_dict)
            air_temperatures.append((min_dict['ta'], max_dict['ta']))
        return air_temperatures

    def evaluate_inside(self, left, right, polygon_name=None):
        """Get a data collection for polygon inclusion from left and right polylines.

        This will be a single 0 or 1 if there is only one value plotted on the chart.
        
        Args:
            left: A Polyline2D for the left of the polygon.
            right: A Polyline2D for the right of the polygon.
            polygon_name: An optional name to be used to create to the data
                collection metadata.
        """
        value_list = []
        vec = Vector2D(1, 0)
        for pt in self._psychrometric_chart.data_points:
            ray = Ray2D(pt, vec)
            if len(right.intersect_line_ray(ray)) != 0:
                if len(left.intersect_line_ray(ray)) == 0:
                    value_list.append(1)
                else:
                    value_list.append(0)
            else:
                value_list.append(0)
        psy = self.psychrometric_chart
        base = psy.temperature if isinstance(psy.temperature, BaseCollection) \
            else psy.relative_humidity
        if isinstance(base, (float, int)):
            return value_list[0]
        coll = base.get_aligned_collection(
            value_list, self.POLYGON_INCLUSION_TYPE, 'status')
        if polygon_name:
            coll.header.metadata = {'polygon': polygon_name}
        return coll

    def create_collection(self, value_list, polygon_name=None):
        """Create a data collection of comfort data values from a list of values.

        Args:
            value_list: A list of data that align with the number of points in the
                underlying psychrometric chart
            polygon_name: An optional name to be used to create to the data
                collection metadata.
        """
        psy = self.psychrometric_chart
        base = psy.temperature if isinstance(psy.temperature, BaseCollection) \
            else psy.relative_humidity
        coll = base.get_aligned_collection(value_list, ThermalComfort(), 'condition')
        if polygon_name:
            coll.header.metadata = {'polygon': polygon_name}
        return coll

    def _build_comfort_polygon(self, left, right):
        """Build a comfort polygon from left and right polylines."""
        # create the saturation line
        psy = self.psychrometric_chart
        x_mid = (left[-1].x + right[-1].x) / 2
        t_mid, t_mid_c = self._x_to_t(x_mid)
        hr_mid = humid_ratio_from_db_rh(t_mid_c, 100, psy.average_pressure)
        mx, my = psy.t_x_value(t_mid), psy.hr_y_value(hr_mid)
        sat_line = Polyline2D((left[-1], Point2D(mx, my), right[-1]),
                                interpolated=True)

        # create the comfort polygon
        comf_polygon = [left.reverse()]
        comf_polygon.append(LineSegment2D.from_end_points(left[0], right[0]))
        comf_polygon.append(right)
        comf_polygon.append(sat_line)
        return tuple(comf_polygon)

    def _evaluate_comfort(self, left, right):
        """Get a tuple of 0s and 1s for comfort from left and right polylines."""
        comfort_vals = []
        vec = Vector2D(1, 0)
        for pt in self._psychrometric_chart.data_points:
            ray = Ray2D(pt, vec)
            if len(right.intersect_line_ray(ray)) != 0:
                if len(left.intersect_line_ray(ray)) == 0:
                    comfort_vals.append(1)
                else:
                    comfort_vals.append(0)
            else:
                comfort_vals.append(0)
        return tuple(comfort_vals)

    def _utci_dict(self, polygon_index):
        """Get a UTCI dictionary for on set of inputs."""
        return {
            'ta': None,
            'tr': self._rad_temperature[polygon_index],
            'vel': self._wind_speed[polygon_index]
        }

    def _x_to_t(self, x_value):
        """Convert an X value on the psychrometric chart to a temperature."""
        psy = self.psychrometric_chart
        t_val = ((x_value - psy.base_point.x) / psy.x_dim) + psy.min_temperature
        t_val_c = t_val if not psy.use_ip else \
            self.TEMP_TYPE.to_unit([t_val], 'C', 'F')[0]
        return t_val, t_val_c

    def _y_to_hr(self, y_value):
        """Convert an Y value on the psychrometric chart to a humidity ratio."""
        psy = self.psychrometric_chart
        return (y_value - psy.base_point.y) / psy._y_dim

    def _check_input(self, input_param, input_name, default=None, check_positive=False):
        """Check a given input value."""
        if input_param is not None and len(input_param) != 0:
            assert isinstance(input_param, (list, tuple)), \
                'Input {} must be a list or a tuple.'.format(input_name)
            new_input_param = []
            for val in input_param:
                if val is not None:
                    val = float(val)
                    if check_positive:
                        assert val >= 0, 'Input {} must be greater or equal ' \
                            'to 0.'.format(input_name)
                new_input_param.append(val)
            input_param = tuple(new_input_param)
            if len(input_param) != self._polygon_count:
                return input_param + \
                    (input_param[-1],) * (self._polygon_count - len(input_param))
            return input_param
        else:
            return (default,) * self._polygon_count

    @staticmethod
    def _min_polylines(polylines):
        """Construct a minimum polyline form a list of polylines."""
        vert_list = list(polylines[0].vertices)
        for poly in polylines[1:]:
            for i, vert in enumerate(poly.vertices):
                if vert.x < vert_list[i].x:
                    vert_list[i] = vert
        return Polyline2D(vert_list, interpolated=True)

    @staticmethod
    def _max_polylines(polylines):
        """Construct a maximum polyline form a list of polylines."""
        vert_list = list(polylines[0].vertices)
        for poly in polylines[1:]:
            for i, vert in enumerate(poly.vertices):
                if vert.x > vert_list[i].x:
                    vert_list[i] = vert
        return Polyline2D(vert_list, interpolated=True)

    @staticmethod
    def _min_index(polylines):
        """Get the index of the left-most polygon."""
        x_vals = [p[0].x for p in polylines]
        return x_vals.index(min(x_vals))

    @staticmethod
    def _max_index(polylines):
        """Get the index of the left-most polygon."""
        x_vals = [p[0].x for p in polylines]
        return x_vals.index(max(x_vals))

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __repr__(self):
        """PolygonPMV representation."""
        return "Polygon PMV: ({} Polygons)".format(self._polygon_count)



# coding=utf-8
"""Utility functions for calculating Universal Thermal Climate Index (UTCI).

This module is devoted to calculating UTCI with NumPy.
"""
import numpy as np
from ..utci import _utci_polynomial


def universal_thermal_climate_index_np(ta, tr, vel, rh):
    """Calculate Universal Thermal Climate Index (UTCI) using a polynomial approximation.

    This function is the same as the base universal_thermal_climate_index function
    but it uses NumPy arrays for calculating UTCI.

    Args:
        ta: Air temperature [C] as a NumPy array.
        tr: Mean radiant temperature [C] as a NumPy array.
        vel: Wind speed 10 m above ground level [m/s] as a NumPy array.
            Note that this meteorological speed at 10 m is simply 1.5 times the
            speed felt at ground in the original Fiala model used to build UTCI.
        rh: Relative humidity [%] as a NumPy array.

    Returns:
        UTCI_approx -- The Universal Thermal Climate Index (UTCI) for the input
        conditions as approximated by a 4-D polynomial.
    """
    # set upper and lower limits of air velocity according to Fiala model scenarios
    vel = np.where(vel < 0.5, 0.5, np.where(vel > 17, 17, vel))

    # metrics derived from the inputs used in the polynomial equation
    eh_pa = saturated_vapor_pressure_hpa_np(
        ta) * (rh / 100.0).astype(np.float32)  # partial vapor pressure
    pa_pr = eh_pa / 10.0  # convert vapour pressure to kPa
    d_tr = tr - ta  # difference between radiant and air temperature

    # pre-calculate powers so we can re-use them
    utci_approx = _utci_polynomial(ta, d_tr, vel, pa_pr)

    return utci_approx


def saturated_vapor_pressure_hpa_np(db_temp):
    """Calculate saturated vapor pressure (hPa) at temperature (C).

    This equation of saturation vapor pressure is specific to the UTCI model.
    """
    g = np.array([-2836.5744, -6028.076559, 19.54263612, -0.02737830188,
                  0.000016261698, 7.0229056e-10, -1.8680009e-13])
    tk = db_temp + 273.15  # air temp in K
    es = 2.7150305 * np.log(tk)
    for i, x in enumerate(g):
        es = es + (x * (tk**(i - 2)))
    es = np.exp(es) * 0.01

    return es


def thermal_condition_np(utci, comfort_par):
    """Determine whether conditions are cold, neutral or hot.

    Values are one of the following:

    * -1 = cold
    * 0 = netural
    * +1 = hot
    """
    conditions = [
        utci < comfort_par.cold_thresh,
        utci > comfort_par.heat_thresh
    ]
    choices = [-1, 1]
    result = np.select(conditions, choices, default=0)

    return result

def thermal_condition_eleven_point_np(utci, comfort_par):
    """Determine the thermal condition on an eleven-point scale.

    Values are one of the following:

    * -5 = extreme cold stress
    * -4 = very strong cold stress
    * -3 = strong cold stress
    * -2 = moderate cold stress
    * -1 = slight cold stress
    * 0 = no thermal stress
    * +1 = slight heat stress
    * +2 = moderate heat stress
    * +3 = strong heat stress
    * +4 = very strong heat stress
    * +5 = extreme heat stress
    """
    conditions = [
        utci < comfort_par.extreme_cold_thresh,
        (utci >= comfort_par.extreme_cold_thresh) & (utci < comfort_par.very_strong_cold_thresh),
        (utci >= comfort_par.very_strong_cold_thresh) & (utci < comfort_par.strong_cold_thresh),
        (utci >= comfort_par.strong_cold_thresh) & (utci < comfort_par.moderate_cold_thresh),
        (utci >= comfort_par.moderate_cold_thresh) & (utci < comfort_par.cold_thresh),
        utci > comfort_par.extreme_heat_thresh,
        (utci <= comfort_par.extreme_heat_thresh) & (utci > comfort_par.very_strong_heat_thresh),
        (utci <= comfort_par.very_strong_heat_thresh) & (utci > comfort_par.strong_heat_thresh),
        (utci <= comfort_par.strong_heat_thresh) & (utci > comfort_par.moderate_heat_thresh),
        (utci <= comfort_par.moderate_heat_thresh) & (utci > comfort_par.heat_thresh)
    ]

    choices = [-5, -4, -3, -2, -1, 5, 4, 3, 2, 1]
    result = np.select(conditions, choices, default=0)

    return result



"""ladybug-comfort commands."""
import click

from ladybug.cli import main
from .epw import epw
from .sql import sql
from .map import map
from .mtx import mtx
from .datacollection import datacollection


# command group for all comfort extension commands.
@click.group(help='ladybug comfort commands.')
@click.version_option()
def comfort():
    pass


# add sub-groups for comfort
comfort.add_command(epw)
comfort.add_command(sql)
comfort.add_command(map)
comfort.add_command(mtx)
comfort.add_command(datacollection, name='data-collection')


# add comfort sub-group to ladybug CLI
main.add_command(comfort)



# coding=utf-8
"""Process comfort map results into thermal comfort percent."""
from __future__ import division
import json

from ._helper import load_matrix


def tcp_model_schedules(
        condition_csv, enclosure_info_json, occ_schedule_json, outdoor_occ_csv=None):
    """Compute Thermal Comfort Percent (TCP) using model-exported occupancy schedules.

    Args:
        condition_csv: Path to a CSV file of thermal conditions output by a
            thermal mapping command.
        enclosure_info_json: Path to a JSON file containing information about
            the radiant enclosure that sensor points belong to. Note that this
            enclosure JSON should be for the same grid as the condition_csv.
        occ_schedule_json: Path to an occupancy schedule JSON output by the
            honeybee-energy model-occ-schedules command.
        outdoor_occ_csv: An optional path to a CSV file to specify the hours during
            which the outdoors is occupied. If None, it will be assumed that
            all hours on the outdoors are occupied. (Default: None).

    Returns:
        A tuple fo three values

        * tcp_list -- List of Thermal Comfort Percent (TCP) values for each sensor.

        * hsp_list -- List of Heat Sensation Percent (HSP) values for each sensor.

        * csp_list - List of Cold Sensation Percent (CSP) values for each sensor.
    """
    # parse all of the input files
    cond_mtx = load_matrix(condition_csv)

    with open(enclosure_info_json) as json_file:
        enclosure_dict = json.load(json_file)
    with open(occ_schedule_json) as json_file:
        occ_dict = json.load(json_file)

    # order the occ schedule data based on the relevant zones from the enclosure_info
    occ_values, total_occs, time_count = [], [], len(cond_mtx[0])
    for zone_id in enclosure_dict['mapper']:
        sch_id = occ_dict['room_occupancy'][zone_id]
        if sch_id is not None:
            sch_vals = occ_dict['schedules'][sch_id]
            occ_values.append(sch_vals)
            total_occs.append(sum(sch_vals))
        else:
            occ_values.append(None)
            total_occs.append(None)
    if enclosure_dict['has_outdoor']:
        if outdoor_occ_csv is None:  # assume the outdoors is always occupied
            sch_vals = [1] * time_count
            occ_values.append(sch_vals)
            total_occs.append(time_count)
        else:
            with open(outdoor_occ_csv) as hourly_schedule:
                sch_vals = [int(float(v)) for v in hourly_schedule]
            assert len(sch_vals) == time_count, 'The number of values in the ' \
                'outdoor occupancy schedule does not match the number of hours ' \
                'for which the thermal map was run.'
            occ_values.append(sch_vals)
            total_occs.append(sum(sch_vals))

    # loop through the sensors and compute tcp, hsp, and csp
    tcp_list, hsp_list, csp_list = [], [], []
    for pt_i, conditions in zip(enclosure_dict['sensor_indices'], cond_mtx):
        occ_sch, total_occ = occ_values[pt_i], total_occs[pt_i]
        if occ_sch is not None and total_occ != 0:
            tcp, hsp, csp = 0, 0, 0
            for occ, cond in zip(occ_sch, conditions):
                if occ == 1:
                    if cond == 0:
                        tcp += 1
                    elif cond == 1:
                        hsp += 1
                    else:
                        csp += 1
            tcp_list.append((tcp / total_occ) * 100)
            hsp_list.append((hsp / total_occ) * 100)
            csp_list.append((csp / total_occ) * 100)
        else:  # the space is unoccupied; treat all hours as relevant
            tcp, hsp, csp = 0, 0, 0
            for cond in conditions:
                if cond == 0:
                    tcp += 1
                elif cond == 1:
                    hsp += 1
                else:
                    csp += 1
            tcp_list.append((tcp / time_count) * 100)
            hsp_list.append((hsp / time_count) * 100)
            csp_list.append((csp / time_count) * 100)
    return tcp_list, hsp_list, csp_list


def tcp_total(condition_csv, schedule=None):
    """Compute Thermal Comfort Percent (TCP) assuming all times are occupied.

    Args:
        condition_csv: Path to a CSV file of thermal conditions output by a
            thermal mapping command.
        schedule: An optional path to a CSV file to specify the relevant times
            during which comfort should be evaluated. If None, it will be assumed that
            all hours are relevant. (Default: None).

    Returns:
        A tuple fo three values.

        * tcp_list -- List of Thermal Comfort Percent (TCP) values for each sensor.

        * hsp_list -- List of Heat Sensation Percent (HSP) values for each sensor.

        * csp_list - List of Cold Sensation Percent (CSP) values for each sensor.
    """
    # parse the csv of results
    cond_mtx = load_matrix(condition_csv)

    # create the occupancy schedule
    time_count = len(cond_mtx[0])
    if schedule is None:
        sch_vals = [1] * time_count
        total_occ = time_count
    else:
        with open(schedule) as hourly_schedule:
            sch_vals = [int(float(v)) for v in hourly_schedule]
        assert len(sch_vals) == time_count, 'The number of values in the ' \
            'outdoor occupancy schedule does not match the number of hours ' \
            'for which the thermal map was run.'
        total_occ = sum(sch_vals)
        assert total_occ != 0, 'No hours of the occupancy schedule are occupied.'

    # loop through the sensors and compute tcp, hsp, and csp
    tcp_list, hsp_list, csp_list = [], [], []
    for conditions in cond_mtx:
        tcp, hsp, csp = 0, 0, 0
        for occ, cond in zip(sch_vals, conditions):
            if occ == 1:
                if cond == 0:
                    tcp += 1
                elif cond == 1:
                    hsp += 1
                else:
                    csp += 1
        tcp_list.append((tcp / total_occ) * 100)
        hsp_list.append((hsp / total_occ) * 100)
        csp_list.append((csp / total_occ) * 100)
    return tcp_list, hsp_list, csp_list



"""A collection of helper functions used throughout the CLI.

Most functions assist with the serialization of objects to/from JSON or CSV.
"""
import os
import json
import numpy as np

from ladybug.analysisperiod import AnalysisPeriod
from ladybug.header import Header
from ladybug.datacollection import HourlyContinuousCollection
from ladybug.futil import preparedir

from ladybug_comfort.parameter.pmv import PMVParameter
from ladybug_comfort.parameter.adaptive import AdaptiveParameter
from ladybug_comfort.parameter.utci import UTCIParameter
from ladybug_comfort.parameter.solarcal import SolarCalParameter


def load_data(values, base_data, data_type, data_units):
    """Load a JSON array string of values to a data collection.

    Args:
        values: A number or JSON array string of numbers.
        base_data: A DataCollection object that serves as the base template
            for re-serialization.
        data_type: The class of the data type for the values.
        data_units: The units of the values.
    """
    if values is not None and values != '' and values != 'None':
        if values.startswith('['):  # it's an array of values
            value_list = json.loads(values)
            header = Header(data_type(), data_units, base_data.header.analysis_period)
            if isinstance(base_data, HourlyContinuousCollection):
                return HourlyContinuousCollection(header, value_list)
            else:
                return base_data.__class__(header, value_list, base_data.datetimes)
        else:  # assume the user has passed a single number
            return float(values)


def load_values(values):
    """Load a single value or JSON array string of values.

    Args:
        values: A number or JSON array string of numbers.
    """
    if values is not None and values != '' and values != 'None':
        if values.startswith('['):  # it's an array of values
            return json.loads(values)
        else:  # assume the user has passed a single number
            return float(values)


def load_value_list(values, calc_len, default=None):
    """Load a single value or JSON array string of values to a list of values

    Args:
        values: A number or JSON array string of numbers.
        calc_len: Integer for the length of the list to be returned.
        default: Default value to be used when the values are None.
    """
    if values is not None and values != '' and values != 'None':
        if values.startswith('['):  # it's an array of values
            return json.loads(values)
        elif os.path.isfile(values):  # it's a CSV with the values in it
            with open(values) as hourly_schedule:
                vals = [float(v) for v in hourly_schedule]
            if len(vals) == 1:
                return vals * calc_len
            return vals
        else:  # assume the user has passed a single number
            try:
                return [float(values)] * calc_len
            except ValueError:  # none of the above; just revert to the default
                pass
    return [default] * calc_len


def load_analysis_period_str(analysis_period_str):
    """Load an AnalysisPeriod from a string.

    Args:
        analysis_period_str: A string of an AnalysisPeriod to be loaded.
    """
    if analysis_period_str is not None and analysis_period_str != '' \
            and analysis_period_str != 'None':
        return AnalysisPeriod.from_string(analysis_period_str)


def load_pmv_par_str(comfort_par_str):
    """Load a PMVParameter from a string.

    Args:
        comfort_par_str: A string of a PMVParameter to be loaded.
    """
    if comfort_par_str is not None and comfort_par_str != '' \
            and comfort_par_str != 'None':
        return PMVParameter.from_string(comfort_par_str)
    return PMVParameter()


def load_adaptive_par_str(comfort_par_str):
    """Load a AdaptiveParameter from a string.

    Args:
        comfort_par_str: A string of a AdaptiveParameter to be loaded.
    """
    if comfort_par_str is not None and comfort_par_str != '' \
            and comfort_par_str != 'None':
        return AdaptiveParameter.from_string(comfort_par_str)
    return AdaptiveParameter()


def load_utci_par_str(comfort_par_str):
    """Load a UTCIParameter from a string.

    Args:
        comfort_par_str: A string of a UTCIParameter to be loaded.
    """
    if comfort_par_str is not None and comfort_par_str != '' \
            and comfort_par_str != 'None':
        return UTCIParameter.from_string(comfort_par_str)
    return UTCIParameter()


def load_solarcal_par_str(solarcal_par_str):
    """Load a SolarCalParameter from a string.

    Args:
        solarcal_par_str: A string of a SolarCalParameter to be loaded.
    """
    if solarcal_par_str is not None and solarcal_par_str != '' \
            and solarcal_par_str != 'None':
        return SolarCalParameter.from_string(solarcal_par_str)


def csv_to_num_matrix(csv_file_path):
    """Load a CSV file consisting only of numbers into a Python matrix of floats.

    Args:
        csv_file_path: Full path to a valid CSV file (e.g. c:/ladybug/test.csv)
    """
    with open(csv_file_path) as csv_data_file:
        return tuple(
            tuple(float(val) for val in row.split(',')) for row in csv_data_file
        )


def _data_to_csv(data, csv_path):
    """Write a list of data collections into a CSV file."""
    with open(csv_path, 'w') as csv_file:
        for dat in data:
            str_data = (str(v) for v in dat)
            csv_file.write(','.join(str_data) + '\n')


def _data_to_ill(data, ill_path):
    """Write a list of data collections into an ill file."""
    with open(ill_path, 'w') as ill_file:
        for dat in data:
            str_data = ('{:.7e}'.format(v) for v in dat)
            ill_file.write(' '.join(str_data) + '\n')


def thermal_map_csv(folder, temperature, condition, condition_intensity,
                    plain_text=True):
    """Write out the thermal mapping CSV files associated with every comfort map."""
    preparedir(folder, remove_content=False)
    result_file_dict = {
        'temperature': os.path.join(folder, 'temperature.csv'),
        'condition': os.path.join(folder, 'condition.csv'),
        'condition_intensity': os.path.join(folder, 'condition_intensity.csv')
    }
    if plain_text:
        _data_to_csv(temperature, result_file_dict['temperature'])
        _data_to_csv(condition, result_file_dict['condition'])
        _data_to_csv(condition_intensity, result_file_dict['condition_intensity'])
    else:
        with open(result_file_dict['temperature'], 'wb') as fp:
            np.save(fp, set_smallest_dtype(np.array(temperature)))
        with open(result_file_dict['condition'], 'wb') as fp:
            np.save(fp, set_smallest_dtype(np.array(condition)))
        with open(result_file_dict['condition_intensity'], 'wb') as fp:
            np.save(fp, set_smallest_dtype(np.array(condition_intensity)))
    return result_file_dict


def smallest_integer_dtype(array: np.ndarray):
    """Return the smallest possible integer dtype.

    Args:
        array: NumPy array.

    Returns:
        A NumPy integer dtype.
    """
    if np.all(array >= np.iinfo(np.int8).min) and \
            np.all(array <= np.iinfo(np.int8).max):
        return np.int8
    elif np.all(array >= np.iinfo(np.int16).min) and \
            np.all(array <= np.iinfo(np.int16).max):
        return np.int16
    elif np.all(array >= np.iinfo(np.int32).min) and \
            np.all(array <= np.iinfo(np.int32).max):
        return np.int32
    elif np.all(array >= np.iinfo(np.int64).min) and \
            np.all(array <= np.iinfo(np.int64).max):
        return np.int64


def smallest_float_dtype(array: np.ndarray, rtol: float = 1e-5, atol: float = 1e-5):
    """Return the smallest possible float dtype.

    The allclose function is used to check if a certain floating-point precision
    can be used without losing accuracy.

    Args:
        array: NumPy array.
        rtol: The relative tolerance parameter for `np.allclose`. The default
            is 1e-5.
        atol: The absolute tolerance parameter for `np.allclose`. The default
            is 1e-5.

    Returns:
        A NumPy floating dtype.
    """
    if np.all((array >= np.finfo(np.float16).min) &
              (array <= np.finfo(np.float16).max)):
        if np.allclose(array, array.astype(np.float16), rtol=rtol, atol=atol):
            return np.float16
    if np.all((array >= np.finfo(np.float32).min) &
              (array <= np.finfo(np.float32).max)):
        if np.allclose(array, array.astype(np.float32), rtol=rtol, atol=atol):
            return np.float32
    if np.all((array >= np.finfo(np.float64).min) &
              (array <= np.finfo(np.float64).max)):
        if np.allclose(array, array.astype(np.float64), rtol=rtol, atol=atol):
            return np.float64


def smallest_dtype(array: np.ndarray, rtol: float = 1e-5, atol: float = 1e-5):
    """Return the smallest possible dtype.

    Args:
        array: NumPy array.
        rtol: The relative tolerance parameter for `np.allclose`. The default
            is 1e-5. This is also used if the dtype of the array is np.floating.
        atol: The absolute tolerance parameter for `np.allclose`. The default
            is 1e-5. This is also used if the dtype of the array is np.floating.

    Returns:
        A NumPy dtype.
    """
    if np.issubdtype(array.dtype, np.integer):
        return smallest_integer_dtype(array)
    elif np.issubdtype(array.dtype, np.floating):
        return smallest_float_dtype(array, rtol=rtol, atol=atol)
    else:
        raise TypeError(f'Expected integer or floating dtype. Got {array.dtype}')


def set_smallest_dtype(array: np.ndarray, rtol: float = 1e-5, atol: float = 1e-5):
    """Return a NumPy array with the smallest possible dtype.

    Args:
        array: NumPy array.
        rtol: The relative tolerance parameter for `np.allclose`. The default
            is 1e-5. This is also used if the dtype of the array is np.floating.
        atol: The absolute tolerance parameter for `np.allclose`. The default
            is 1e-5. This is also used if the dtype of the array is np.floating.

    Returns:
        A new NumPy array with a smaller dtype.
    """
    dtype = smallest_dtype(array, rtol=rtol, atol=atol)
    return array.astype(dtype)



"""Commands to work with data collections."""
import sys
import logging
import numpy as np
from pathlib import Path
import click
import json

from ladybug.datacollection import HourlyContinuousCollection, HourlyDiscontinuousCollection
from ladybug.header import Header
from ladybug.datautil import collections_to_csv


_logger = logging.getLogger(__name__)


@click.group(help='Commands to work with data collections.')
def datacollection():
    pass


@datacollection.command('folder-to-datacollections')
@click.argument(
    'folder', type=click.Path(exists=True, dir_okay=True, resolve_path=True)
)
@click.argument(
    'data-type', type=click.Path(exists=True, dir_okay=False, resolve_path=True)
)
@click.option(
    '--sub-folder', '-sf', type=click.STRING, default='datacollections',
    show_default=True
)
def folder_to_datacollections(folder, data_type, sub_folder):
    """Read an npy file and convert every row to a data collection.

    The data collections will be saved as CSV files in subfolder.

    \b
    Args:
        npy-file: Path to npy file.
        data-type: A JSON file with the data type.
        grid-name: The name of the grid. This is used in the metadata of the
            header.
    """
    with open(Path(folder, 'grids_info.json')) as json_file:
        grid_list = json.load(json_file)
    with open(data_type) as json_file:
        data_header = Header.from_dict(json.load(json_file))
    a_per = data_header.analysis_period
    continuous = True if a_per.st_hour == 0 and a_per.end_hour == 23 else False
    if not continuous:
        dates = a_per.datetimes
    try:
        for grid in grid_list:
            grid_name = grid['full_id'] if 'full_id' in grid else 'id'
            metadata = {'grid': grid_name}
            grid_file = Path(folder, '{}.npy'.format(grid_name))
            data_matrix = np.load(grid_file).tolist()
            grid_data = []
            for i, row in enumerate(data_matrix):
                header = data_header.duplicate()
                header.metadata = metadata.copy()
                header.metadata['sensor_index'] = i
                data = HourlyContinuousCollection(header, row) if continuous else \
                    HourlyDiscontinuousCollection(header, row, dates)
                grid_data.append(data)

            file_name = grid_name + '.csv'
            collections_to_csv(grid_data, Path(folder, sub_folder), file_name)
    except Exception:
        _logger.exception('Failed to convert folder of files to data collections.')
        sys.exit(1)
    else:
        sys.exit(0)



# coding=utf-8
"""Utility functions for computing local thermal discomfort according to ASHRAE-55."""
from __future__ import division

import math


def radiant_asymmetry_ppd(radiant_temperature_difference, asymmetry_type):
    """Calculate the percentage of people dissatisfied from radiant asymmetry.

    The comfort functions used here come from Figure 5.2.4.1 of ASHRAE 55 2010.
    Note that, if the resulting input results in a PPD beyond what is included
    in this Figure, the maximum PPD will simply be returned.

    Note:
        [1] ASHRAE Standard 55 (2010). "Thermal Environmental Conditions
        for Human Occupancy." American Society of Heating,
        Refrigerating and Air Conditioning Engineers.

    Args:
        radiant_temperature_difference: The radiant temperature difference between
            two sides of the same plane where an occupant is located [C].
        asymmetry_type: An integer that representing the type of radiant
            asymmetry being evaluated. Occupants are more sensitive to warm
            ceilings and cool walls than cool ceilings and warm walls.
            Choose from the following options.

            * 0 = WarmCeiling
            * 1 = CoolWall
            * 2 = CoolCeiling
            * 3 = WarmWall

    Returns:
        ppd -- The percentage of people dissatisfied (PPD) for the input
        radiant asymmetry.
    """
    td = radiant_temperature_difference
    if asymmetry_type == 0:
        if td > 23:
            ppd = 100 / (1 + math.exp(2.84 - 0.174 * 23)) - 5.5
        else:
            ppd = 100 / (1 + math.exp(2.84 - 0.174 * td)) - 5.5
    elif asymmetry_type == 1:
        if td > 15:
            ppd = 100 / (1 + math.exp(6.61 - 0.345 * 15))
        else:
            ppd = 100 / (1 + math.exp(6.61 - 0.345 * td))
    elif asymmetry_type == 2:
        if td > 15:
            ppd = 100 / (1 + math.exp(9.93 - 0.50 * 15))
        else:
            ppd = 100 / (1 + math.exp(9.93 - 0.50 * td))
    elif asymmetry_type == 3:
        if td > 35:
            ppd = 100 / (1 + math.exp(3.72 - 0.052 * 35)) - 3.5
        else:
            ppd = 100 / (1 + math.exp(3.72 - 0.052 * td)) - 3.5
    else:
        raise ValueError(
            'Radiant asymmetry_type "{}" was not recognized.'.format(asymmetry_type))
    return ppd


def ankle_draft_ppd(predicted_mean_vote, air_speed_at_ankle):
    """Calculate the percentage of people dissatisfied from cold drafts at ankle-level.

    The original tests used to create the model involved blowing cold air on
    subject's ankles at a height of 10 cm off of the ground.
    The formula was officially incorporated in the ASHRAE 55 standard in 2020
    with a recommendation that PPD from ankle draft not exceed 20%.

    Note:
        [1] ASHRAE Standard 55 (2020). "Thermal Environmental Conditions
        for Human Occupancy." American Society of Heating,
        Refrigerating and Air Conditioning Engineers.

        [2] Schiavon, S., D. Rim, W. Pasut, W. Nazaroff. 2016. "Sensation of
        draft at uncovered ankles for women exposed to displacement ventilation
        and underfloor air distribution systems." Building and Environment, 96, 228-236.

        [3] Liu, S., S. Schiavon, A. Kabanshi, W. Nazaroff. 2016. "Predicted
        percentage of dissatisfied with ankle draft." Accepted Author Manuscript.
        Indoor Environmental Quality. http://escholarship.org/uc/item/9076254n

    Args:
        predicted_mean_vote: The full-body predicted mean vote (PMV) of the subject.
            Ankle draft depends on full-body PMV because subjects are more likely
            to feel uncomfortably cold at their extremities when their whole body
            is already feeling colder than neutral.
            The functions in the pmv module can be used to calculate this input.
        air_speed_at_ankle: The velocity of the draft in m/s at ankle level
            (10cm above the floor).

    Returns:
        ppd -- The percentage of people dissatisfied (PPD) from cold downdraft
        at ankle-level.
    """
    c_term = math.exp(-2.58 + (3.05 * air_speed_at_ankle) - (1.06 * predicted_mean_vote))
    return 100 * (c_term / (1 + c_term))



# coding=utf-8
"""Utility functions for calculating Physiologic Equivalent Temperature (PET).

PET uses the Munich Energy Balance Model (MEMI), which is arguably the most
detailed 3-node human energy balance model in common use today. It can account
for various physiological features of the human subject, including age, sex,
height, and body mass, making it one of the only models that is suitable for
forecasting the thermal experience of a specific individual. This also makes
it one of the better models for estimating core body temperature and whether
a given set of conditions is likely to induce hypothermia or hyperthermia in
a specific individual.

PET is, itself, a "feels like" temperature and is defined as the operative
temperature of a reference environment that would cause the same physiological
response in the human subject as the environment under study. That is, the
same skin temperature and core body temperature.
"""
from __future__ import division
import math

from ladybug.rootfinding import secant_three_var

# constants for standard conditions
TC_SET = 36.6  # standard core body temperature
TSK_SET = 34  # standard skin temperature
TBODY_SET = 0.1 * TSK_SET + 0.9 * TC_SET  # weighted average for full-body temperature
# universal constants
P_REF = 101325  # sea-level barometric pressure [Pa]
C_AIR = 1010  # air specific heat capacity [J/kg-K]
L_VAP = 2420000  # latent heat of evaporation [J/Kg]
C_B = 3640  # blood specific heat [J/kg-K]
EM_SK = 0.99  # human skin emissivity
EM_CL = 0.95  # clothing emissivity
SIGM = 5.67 * (10 ** -8)  # Stefan-Boltzmann constant [W/(m2-K^(-4))]


def physiologic_equivalent_temperature(
        ta, tr, vel, rh, met, clo, age=36, sex=0.5, ht=1.65, m_body=62, pos='standing',
        b_press=101325):
    """Calculate Physiological Equivalent Temperature (PET).

    This method is based on Peter Hoeppe's original PET Fortran code, from the
    VDI Norm 3787, Blatt 2 [1][2], as well as Djordje Spasic's Ladybug Legacy
    source code. It includes the corrected reference environment and updated
    vapor transfer model published by Edouard Walther (AREP, France) and
    Quentin Goestchel (ENS Paris-Saclay, France) [3].

    Chris Mackey added support for the specification of metabolic rates in met
    (instead of Watts above basal metabolism) using a USDA report that relates
    Resting Metabolic Rate (RMR) (assumed to be 1 met) to Basal Metabolic
    Rate (BMR) using a Physical Activity Level (PAL) ratio [4].

    Note:
        [1] Hoeppe, Peter. (1999). "The physiological equivalent temperature - A
        universal index for the biometeorological assessment of the thermal
        environment." International journal of biometeorology. 43. 71-5.
        10.1007/s004840050118.
        https://link.springer.com/article/10.1007/s004840050118

        [2] Hoeppe, Peter. (2008). "Urban climatic map and standards for wind
        environment - Feasibility study, Technical Input Report No.1", The
        Chinese University of Hong Kong, Planning Department.

        [3] Walther, Edouard and Goestchel, Quentin. (2018). "The P.E.T. comfort
        index: Questioning the model". Building and Environment. 137C. 1-10.
        10.1016/j.buildenv.2018.03.054.
        https://www.sciencedirect.com/science/article/pii/S0360132318301896

        [4] "Dietary Reference Intakes for Energy, Carbohydrate, Fiber, Fat,
        Fatty Acids, Cholesterol, Protein, and Amino Acids (Macronutrients) (2005)".
        USDA. National Academy of Sciences, Institute of Medicine, Food and
        Nutrition Board. Archived from the original on 10 March 2016.
        Chapter 12, page 8.
        https://web.archive.org/web/20160310031719/http://fnic.nal.usda.gov/\
        dietary-guidance/dri-nutrient-reports/energy-carbohydrate-fiber-fat-\
        fatty-acids-cholesterol-protein

    Args:
        ta: Air temperature [C].
        tr: Mean radiant temperature [C].
        vel: Relative air velocity [m/s].
        rh: Relative humidity [%].
        met: Metabolic rate [met]. Note that the original PET model requires that
            the activity of the human subject be accounted for as additional Watts
            above the basal metabolism, which is often difficult to estimate.
            In order to accept an input in [met], it is assumed that 1 met refers
            to Resting Metabolic Rate (RMR) and this is 1.17 times the male
            Basal Metabolic Rate (BMR) or 1.22 times the female BMR, where 1.17 =
            1 / 0.9 (TEF or food digestion) / 0.95 (basal to resting activity difference)
        clo: Clothing [clo].
        age: The age of the human subject in years. (Default: 36 years for middle age
            of the average worldwide life expectancy).
        sex: A value between 0 and 1 to indicate the sex of the human subject,
            which influences the computation of basal metabolism. 0 indicates male.
            1 indicates female and any number in between denotes a weighted average
            between the two. (Default: 0.5).
        ht: The height of the human subject in meters. (Default: 1.65m for
            a worldwide average between male and female height).
        m_body: The body mass of the human subject in kilograms. (Default: 62 kg
            for the worldwide average adult human body mass).
        pos: Text to indicate the posture of the human subject's body. Choose from
            the following: "standing", "seated", "crouching". (Default: "standing").
        b_press: An optional number for the air pressure in which the human subject
            exists, which determines the rate of sweat evaporation [Pa]. Default
            is pressure at sea level (101325 Pa).

    Returns:
        A dictionary containing results of the PET model with the following keys

        -   pet -- Physiological equivalent temperature (PET) [C]
        -   t_core -- Core body temperature [C]
        -   t_skin -- Skin temperature [C]
        -   t_clo -- Clothing temperature [C]
    """
    epsilon = 0.01  # the acceptable error in the result of the temperatures

    # compute the constant variables of the MEMI model
    a_du, a_clo, a_effr, feff, hc, fcl, facl, rcl, htcl, vpa, he, ere = \
        _memi_constant_vars(ta, vel, rh, b_press, met, clo, age, sex, ht, m_body, pos)

    # determine a starting guess for the human temperatures using constant variables
    t_core_in = 36.6  # normal human body temperature
    t_env = (ta + tr) / 2  # the operative temperature of the surrounding environment
    r_body = (1 / htcl) - rcl
    r_tot = r_body + rcl
    t_sk_in = t_core_in * (rcl / r_tot) + t_env * (r_body / r_tot)
    t_clo_in = (ta + tr + t_sk_in) / 3  # average of air, radiant, and skin temperature

    # find a steady state solution to the MEMI model balance under the input conditions
    d_args = (ta, tr, a_du, a_clo, a_effr, feff, hc, fcl, facl, rcl, htcl, vpa, he, ere)
    for i in (1, 2, 3, 4, 5, 6):  # progressively widen search range
        t_min = (t_core_in - (5 * i), t_sk_in - (10 * i), t_clo_in - (10 * i))
        t_max = (t_core_in + (5 * i), t_sk_in + (10 * i), t_clo_in + (10 * i))
        try:
            tn = secant_three_var(
                t_min, t_max, _memi_dynamic_balance, epsilon, other_args=d_args)
        except OverflowError:  # about 1% of the time, the solution diverges
            tn = None
        if tn is not None:
            break  # model converged and root was found
    if tn is None:  # if we still don't have convergence, try brute force method
        starting_guess = (t_core_in, t_sk_in, t_clo_in)
        increments = (0.001, 0.01, 0.01)  # increments with which to adjust body temp
        err = 0.1  # maximum allowed error in [W]
        tn = _brute_force_three_var(
            starting_guess, increments, _memi_dynamic_balance, err, other_args=d_args)

    # compute the PET using the human subject temperatures using a bisection method
    def f(tx):
        """A function with the input variables of the PET reference situation."""
        return memi_balance(
            tn, tx, tx, 0.1, 50, met, 0.9, age, sex, ht, m_body, pos, b_press,
            False, True)
    ti = -40  # start of the search interval
    tf = 60  # end of the search interval
    pet = 0
    while tf - ti > epsilon:  # bisection loop
        if f(ti) * f(pet) < 0:
            tf = pet
        else:
            ti = pet
        pet = (ti + tf) / 2

    # put all of the results into a single dictionary
    return {'pet': pet, 't_core': tn[0], 't_skin': tn[1], 't_clo': tn[2]}


def memi_balance(
        t_human, ta, tr, vel, rh, met, clo, age, sex, ht, m_body, pos,
        b_press=101325, actual=True, scalar=False):
    """Perform the human load balance using Munich energy balance model (MEMI).

    Args:
        t_human: A list of three values for the temperature of the human subject.

            * core body temperature [C]
            * skin temperature [C]
            * clothing temperature [C]

        ta: Air temperature [C].
        tr: Mean radiant temperature [C].
        vel: Relative air velocity [m/s].
        rh: Relative humidity [%].
        met: Metabolic rate [met].
        clo: Clothing [clo].
        age: The age of the human subject [years].
        sex: A value between 0 and 1 to indicate the sex of the human [0=male, 1=female].
        ht: The height of the human subject in [m].
        m_body: The body mass of the human subject in [kg].
        pos: Text for the posture of the human subject. [standing, seated, crouching].
        b_press: The barometric air pressure [Pa]. (Default: 101325 for sea level).
        actual: A boolean to indicate whether the calculation should be performed
            in the actual environment (True) or the reference environment (False).
        scalar: A boolean for whether the result should be returned as a vector of
            the energy flux across [core, skin, clo] or it should be a single scalar
            energy flux for the entire human subject.

    Returns:
        The energy flux across the entire human subject if scalar is True or a
        vector with energy flux across [core, skin, clo] if scalar is False.
    """
    # compute constant attributes of the model
    a_du, a_clo, a_effr, feff, hc, fcl, facl, rcl, htcl, vpa, he, ere = \
        _memi_constant_vars(
            ta, vel, rh, b_press, met, clo, age, sex, ht, m_body, pos, actual)
    # compute the dynamic load balance of the model
    return _memi_dynamic_balance(
        t_human, ta, tr, a_du, a_clo, a_effr, feff, hc, fcl, facl, rcl, htcl, vpa,
        he, ere, scalar)


def _memi_constant_vars(
        ta, vel, rh, b_press, met, clo, age, sex, ht, m_body, pos, actual=True):
    """Compute variables of the MEMI model that do not change with body temperature.

    This includes things such as the surface are of the human, basal metabolism,
    vapor pressure, convection coefficients, radiant efficiency, etc.

    Args:
        ta: Air temperature [C].
        vel: Relative air velocity [m/s].
        rh: Relative humidity [%].
        b_press: The barometric air pressure [Pa]. (Default: 101325 for sea level).
        met: Metabolic rate [met].
        clo: Clothing [clo].
        age: The age of the human subject [years].
        sex: A value between 0 and 1 to indicate the sex of the human [0=male, 1=female].
        ht: The height of the human subject in [m].
        m_body: The body mass of the human subject in [kg].
        pos: Text for the posture of the human subject. [standing, seated, crouching].
        actual: A boolean to indicate whether the calculation should be performed
            in the actual environment (True) or the reference environment (False).

    Returns:
        A tuple of variables that are constant throughout the computation of the
        MEMI model energy balance.
    """
    # compute tha area parameters of the body
    a_du = 0.203 * m_body ** 0.425 * ht ** 0.725  # Dubois surface area of human subject
    feff = 0.725 if pos in ('standing', 'crouching') else 0.696  # radiant efficiency

    # increase the Burton surface to account for clothing, k = 0.31 for Hoeppe
    fcl = 1 + (0.31 * clo)  # increase heat exchange surface depending on clothing level
    facl = (173.51 * clo - 2.36 - 100.76 * clo * clo + 19.28 * clo ** 3.0) / 100
    a_clo = a_du * facl + a_du * (fcl - 1.0)
    a_effr = a_du * feff  # effective radiative area derived from posture of the subject

    # partial pressure of water depending on relative humidity and air temperature
    if actual:  # the calculation of the actual vapor pressure of inputs
        vpa = rh / 100.0 * 6.105 * math.exp(17.27 * ta / (237.7 + ta))  # [hPa]
    else:  # use reference temperature, humidity and barometric pressure
        vpa = 12  # [hPa] vapour pressure of the standard environment

    # convection coefficient depending on air speed and subject posture
    if pos == 'standing':
        hc = 2.67 + (6.5 * vel ** 0.67)
    elif pos == 'seated':
        hc = 2.26 + (7.42 * vel ** 0.67)
    elif pos == 'crouching':
        hc = 8.6 * (vel ** 0.513)
    # modification of hc with the total air pressure
    hc = hc * (b_press / P_REF) ** 0.55

    # compute base metabolism for men and women in [W]
    r_fem = ht * 100.0 / m_body ** (1.0 / 3.0) - 42.1
    metab_female = 3.19 * m_body ** 0.75 * (1.0 + 0.004 * (30.0 - age) + 0.018 * r_fem)
    r_mal = ht * 100.0 / m_body ** (1.0 / 3.0) - 43.4
    metab_male = 3.45 * m_body ** 0.75 * (1.0 + 0.004 * (30.0 - age) + 0.01 * r_mal)
    # compute the total metabolism, accounting for the activity level
    if actual:  # actual human subject metabolic rate
        mec = (metab_male * 1.17 * met) / a_du  # [W/m2]
        fec = (metab_female * 1.22 * met) / a_du  # [W/m2]
    else:  # reference human subject metabolic rate assumes 80 W of activity level
        mec = (80 + metab_male) / a_du  # [W/m2]
        fec = (80 + metab_female) / a_du  # [W/m2]

    # attribution of internal energy depending on the sex of the subject
    he = ((1 - sex) * mec) + (sex * fec)  # [W/m2]

    # compute the respiratory energy losses from the metabolic rate
    texp = 0.47 * ta + 21.0  # [degC]
    dventpulm = he * 1.44 * (10.0 ** -6)  # pulmonary flow rate
    eres = C_AIR * (ta - texp) * dventpulm  # sensible heat loss [W/m2]
    vpexp = 6.11 * 10.0 ** (7.45 * texp / (235.0 + texp))
    p_hpa = b_press / 100  # barometric pressure [hPa]
    erel = 0.623 * L_VAP / p_hpa * (vpa - vpexp) * dventpulm  # latent heat loss [W/m2]
    ere = eres + erel  # total respiratory heat loss [W/m2]

    # compute the clothed fraction of the body and the clothing thickness
    rcl = clo / 6.45  # convert [clo] to [m2-K/W]
    if facl > 1.0:  # ensure that clothing does not cover more than 100%
        facl = 1.0
    y = 0  # thickness of the clothing layer
    if clo >= 2.0:
        y = 1.0
    elif clo > 0.6:
        y = (ht - 0.2) / ht
    elif clo > 0.3:
        y = 0.5
    elif clo > 0.0:
        y = 0.1

    # compute subject radius depending on the clothing level (6.28 = 2 * pi)
    r2 = a_du * (fcl - 1.0 + facl) / (6.28 * ht * y)  # external radius
    r1 = facl * a_du / (6.28 * ht * y)  # internal radius
    di = r2 - r1

    # compute the equivalent thermal resistance of body tissues
    htcl = (6.28 * ht * y * di) / (rcl * math.log(r2 / r1) * a_clo)  # [W/(m2-K)]

    # return all of the variables used by the rest of the MEMI calculation
    return a_du, a_clo, a_effr, feff, hc, fcl, facl, rcl, htcl, vpa, he, ere


def _memi_dynamic_balance(
        t_human, ta, tr, a_du, a_clo, a_effr, feff, hc, fcl, facl, rcl, htcl, vpa,
        he, ere, scalar=False):
    """Compute the dynamic load balance of the MEMI model.

    Args:
        t_human: A list of three values for the temperature of the human subject.

            * core body temperature [C]
            * skin temperature [C]
            * clothing temperature [C]

        ta: Air temperature [C].
        tr: Mean radiant temperature [C].
        a_du: The Dubois surface area of the human subject [m2].
        a_clo: The clothing surface area of the human subject [m2].
        a_effr: The effective radiant surface area of the human subject [m2].
        feff: The fractional radiant efficiency of the human subject.
        hc: The coefficient of convective thermal transfer [W/m2-k].
        fcl: The fractional increase in heat exchange surface depending on clothing.
        facl: The fraction of the human subject covered in clothing.
        rcl: The thermal resistance of the clothing layer [m2-K/W].
        htcl: The thermal transmittance of all body tissues and clothing [W/(m2-K)].
        vpa: The partial vapour pressure of the environment [hPa].
        he: The internal heat energy generated by the human subject [W/m2].
        ere: The total respiratory heat loss [W/m2].
        scalar: A boolean for whether the result should be returned as a vector of
            the energy flux across [core, skin, clo] or it should be a single scalar
            energy flux for the entire human subject.

    Returns:
        The energy flux across the entire human subject if scalar is True or a
        vector with energy flux across [core, skin, clo] if scalar is False.
    """
    # unpack the array of temperatures of the human subject and get average body temp
    t_core, t_sk, t_clo = t_human
    alpha = 0.1  # constant in steady state model but updates t_body in transient model
    t_body = alpha * t_sk + (1 - alpha) * t_core

    # compute sweat losses
    qmsw = sweat_volume(t_body)
    # L_VAP / 1000 --> [J/g] ; qwsw / 3600 --> [g/m2-s]
    esw = (L_VAP / 1000) * (qmsw / 3600)  # [W/m2]
    # saturation vapor pressure at temperature tsk
    pv_sk = 6.105 * math.exp((17.27 * (t_sk + 273.15) - 4717.03) / (237.7 + t_sk))  # hPa
    # compute vapour transfer
    lw = 1.67  # Lewis factor [K/hPa]
    he_diff = hc * lw  # diffusion coefficient of air layer
    fecl = 1 / (1 + 0.92 * hc * rcl)  # Burton efficiency factor
    emax = he_diff * fecl * (pv_sk - vpa)  # maximum diffusion at skin surface
    w = esw / emax  # skin wettedness
    if w > 1:
        w = 1
        delta = esw - emax
        if delta < 0:
            esw = emax
    if esw < 0:
        esw = 0
    i_m = 0.38  # Woodcock's ratio
    r_ecl = (1 / (fcl * hc) + rcl) / (lw * i_m)  # clothing vapour transfer resistance
    ediff = (1 - w) * (pv_sk - vpa) / r_ecl  # diffusion heat transfer
    evap = -(ediff + esw)  # [W/m2]

    # compute radiation losses
    tr_k, tsk_k, tclo_k = tr + 273.15, t_sk + 273.15, t_clo + 273.15
    # for bare skin area [W/m2]
    rbare = a_effr * (1.0 - facl) * EM_SK * SIGM * (tr_k ** 4 - tsk_k ** 4) / a_du
    # for dressed area [W/m2]
    rclo = feff * a_clo * EM_CL * SIGM * (tr_k ** 4 - tclo_k ** 4) / a_du

    # compute convection losses #
    # for bare skin area:
    cbare = hc * (ta - t_sk) * a_du * (1.0 - facl) / a_du  # [W/m2]
    # for dressed area:
    cclo = hc * (ta - t_clo) * a_clo / a_du  # [W/m2]

    # return either the calculated [core, skin, clo] energy flux or the scalar sum
    if scalar:  # return the scalar sum of the energy balance
        return he + ere + rclo + rbare + cclo + cbare + evap
    else:  # produce a vector of 3 energy fluxes across [core, skin, clo]
        vaso_ex = (vaso_circulation(t_core, t_sk) / 3600 * C_B + 5.28) * (t_core - t_sk)
        clo_ex = htcl * (t_sk - t_clo)
        e_core = he + ere - vaso_ex  # core balance [W/m2]
        e_sk = rbare + cbare + evap + vaso_ex - clo_ex  # skin balance [W/m2]
        e_clo = cclo + rclo + clo_ex  # clothes balance [W/m2]
        return (e_core, e_sk, e_clo)


def vaso_circulation(t_core, t_skin):
    """Calculate skin blood flow (vaso-circulation) using core and skin temperatures.

    Args:
        t_core: The core temperature of the human subject. [C].
        t_skin: The skin temperature of the human subject. [C].

    Returns:
         A number for the volume of blood flow in L/m2-h.
    """
    # set value signals
    sig_skin = TSK_SET - t_skin
    sig_core = t_core - TC_SET
    if sig_core < 0:  # t_core < TC_SET --> the blood flow is reduced
        sig_core = 0
    if sig_skin < 0:  # t_skin > TSK_SET --> the blood flow is increased
        sig_skin = 0
    # 6.3 L/m2-h is the standard volume of the blood flow
    qm_blood = (6.3 + 75 * sig_core) / (1 + 0.5 * sig_skin)
    return qm_blood if qm_blood < 90 else 90  # 90 L/m2-h is the blood flow upper limit


def sweat_volume(t_body):
    """Calculate skin sweat volume using average body temperatures.

    Args:
        t_body: The average body temperature of the human subject. [C].

    Returns:
         A number for the volume of sweat in g/m2-h.
    """
    sig_body = t_body - TBODY_SET
    if sig_body < 0:  # Tbody < Tbody_set --> The sweat flow is 0
        sig_body = 0
    qm_sw = 304.94 * 10 ** -3 * sig_body
    return qm_sw if qm_sw < 500 else 500  # 500 g/m2-h is the upper limit of sweat rate


def pet_category(pet):
    """Get the category of heat/cold stress associated PET.

    This method uses the original categories developed by Matzarakis and Mayer (1996),
    which are most suitable in temperature climates.

    Each number (from -4 to +4) represents a certain PET thermal sensation category.

    * -4 = Extreme Cold
    * -3 = Strong Cold
    * -2 = Moderate Cold
    * -1 = Slight Cold
    * 0 = Comfortable
    * 1 = Slight Heat
    * 2 = Moderate Heat
    * 3 = Strong Heat
    * 4 = Extreme Heat

    Args:
        pet: Physiological Equivalent Temperature [C].

    Returns:
        category -- An integer indicating the level of heat/cold stress
        associated with the PET.
    """
    if pet < 4:
        return -4
    elif pet < 8:
        return -3
    elif pet < 13:
        return -2
    elif pet < 18:
        return -1
    elif pet <= 23:
        return 0
    elif pet <= 29:
        return 1
    elif pet <= 35:
        return 2
    elif pet <= 41:
        return 3
    else:
        return 4


def pet_category_humid(pet):
    """Get the category of heat/cold stress associated PET.

    This method uses the categories by Lin and Matzarakis (2008), which are
    suitable for tropical and subtropical humid climates.

    Each number (from -4 to +4) represents a certain PET thermal sensation category.

    * -4 = Extreme Cold
    * -3 = Strong Cold
    * -2 = Moderate Cold
    * -1 = Slight Cold
    * 0 = Comfortable
    * 1 = Slight Heat
    * 2 = Moderate Heat
    * 3 = Strong Heat
    * 4 = Extreme Heat

    Args:
        pet: Physiological Equivalent Temperature [C].

    Returns:
        category -- An integer indicating the level of heat/cold stress
        associated with the PET.
    """
    if pet < 14:
        return -4
    elif pet < 18:
        return -3
    elif pet < 22:
        return -2
    elif pet < 26:
        return -1
    elif pet <= 30:
        return 0
    elif pet <= 34:
        return 1
    elif pet <= 38:
        return 2
    elif pet <= 42:
        return 3
    else:
        return 4


def core_temperature_category(t_core):
    """Get the classification of core body temperature.

    * -2 = Hypothermia
    * -1 = Cold
    * 0 = Normal
    * 1 = Hot
    * 2 = Hyperthermia

    Args:
        t_core: The core body temperature of the human subject [C].

    Returns:
        category -- An integer indicating the classification of the body temperature.

    Note:
        [1] Marx J (2006). Rosen's emergency medicine : concepts and clinical
        practice (6th ed.). Philadelphia: Mosby/Elsevier. p. 2239.
        ISBN 978-0-323-02845-5. OCLC 58533794.
    """
    if t_core < 35:
        return -2
    elif t_core < 36.5:
        return -1
    elif t_core < 37.5:
        return 0
    elif t_core < 38.3:
        return 1
    else:
        return 2


def _brute_force_three_var(
        starting_guess, increments, fn, epsilon, other_args, max_iter=100000):
    """Brute force root-finding algorithm designed specifically for solving MEMI models.

    It is reliable. However, it converges slowly and, for this reason, it is recommended
    that this only be used after the secant_three_var() method has returned None.

    Args:
        starting_guess: A tuple with 3 numbers with a starting guess for the roots.
        increments: A tuple with 3 numbers indicating the increments with which
            each value should be adjusted to find the root.
        fn: A function for which roots are to be solved. That is, where the output
            of the function is a tuple of three zeros.
        epsilon: The acceptable error in the output of the function (aka. how
            far from zero it is allowed to be).
        other_args: Other input arguments for the fn other than the ones being
            adjusted to solve the root.
        max_iter: The maximum number of iterations with which a root will be
            sought. (Default: 100000).

    Returns:
        root -- a tuple of 3 values that return a vector of zeros from the fn.
    """
    curr_i = 0
    in_args = (starting_guess,) + other_args
    en_vec = fn(*in_args)
    while not all(abs(v) < epsilon for v in en_vec):
        new_guess = list(starting_guess)
        for i, hf in enumerate(en_vec):
            if hf > epsilon:
                new_guess[i] = new_guess[i] + increments[i]
            elif hf < -epsilon:
                new_guess[i] = new_guess[i] - increments[i]
        starting_guess = new_guess
        in_args = (starting_guess,) + other_args
        en_vec = fn(*in_args)
        curr_i += 1
        if curr_i == max_iter:
            break
    return starting_guess
